

    /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
    *                                                                 *
    *   DEVIL Screensaver System 1.03                                 *
    *   Copyright  2005 Sebastian Pipping <webmaster@hartwork.org>   *
    *                                                                 *
    *   --> http://www.hartwork.org                                   *
    *                                                                 *
    *                                                                 *
    *   This source code is released under LGPL.                      *
    *   See LGPL.txt for details.                        2005-08-31   *
    *                                                                 *
    \* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


#include "Pool.h"
#include <time.h>
#include <stdlib.h>
#include <math.h>


#define MAX_STIMUS	30
#define POOL_WIDTH	320		// multiple of 4, important for bitmap
#define POOL_HEIGHT	240		// same here

#define TWO_PI ( 2 * 3.1415926535 )


// *****************************************************************************
Pool::Pool()
{
	uWidth  = POOL_WIDTH;
	uHeight = POOL_HEIGHT;
	uParticleCount = POOL_WIDTH * POOL_HEIGHT;

	memset( &bmi.bmiHeader, 0, sizeof( BITMAPINFOHEADER ) );
	bmi.bmiHeader.biSize		= sizeof( BITMAPINFOHEADER );
	bmi.bmiHeader.biWidth		= uWidth;
	bmi.bmiHeader.biHeight		= uHeight;
	bmi.bmiHeader.biPlanes		= 1;
	bmi.bmiHeader.biBitCount	= 24;
	bmi.bmiHeader.biCompression	= BI_RGB;

	image = new RGBTRIPLE[ uParticleCount ];

	waterplane[ 0 ] = new Particle[ uParticleCount ];
	waterplane[ 1 ] = new Particle[ uParticleCount ];
	ZeroMemory( waterplane[ 0 ], uParticleCount * sizeof( Particle ) );
	ZeroMemory( waterplane[ 1 ], uParticleCount * sizeof( Particle ) );

	bWhichPlane = false;
	fSinusSeed = 0.0;
	uStimus = 0;
	first = NULL;

	// Init random
	srand( time( NULL ) );

	// Update full image one time
	UpdateImage( true );
}


// *****************************************************************************
Pool::~Pool()
{
	delete [] image;
	delete [] waterplane[ 0 ];
	delete [] waterplane[ 1 ];

	// Delete Stimu list
	Stimu * temp = NULL;
	while( first != NULL )
	{
		temp = first->next;
		delete first;
		first = temp;
	}
}


// *****************************************************************************
void Pool::CreateStimu( Stimu * out, Stimu * next )
{
	// Get random location
	UINT uIndex = ( UINT )( ( rand() * uParticleCount ) / RAND_MAX );

	out->age		= 7;
	out->intensity	= 120;
	out->x			= uIndex % uWidth;
	out->y			= uIndex / uWidth;
	out->next		= next;
}


// *****************************************************************************
void Pool::AddStimu()
{
	// Insert new Stimu at the beginning
	Stimu * temp = first;
	first = new Stimu;
	CreateStimu( first, temp );

	uStimus++;
}


// *****************************************************************************
void Pool::StepStimus()
{
	fSinusSeed += 0.19;
	while( fSinusSeed > TWO_PI ) fSinusSeed -= TWO_PI;

	UINT uParticleIndex = 0;
	UINT uStimuIndex = 0;

	Stimu * temp = first;
	while( temp )
	{
		if(  --( temp->age ) <= 0 )
		{
			// Replace Stimu
			CreateStimu( temp, temp->next );
		}
		
		uParticleIndex = temp->x + uWidth * temp->y;
		waterplane[ bWhichPlane ][ uParticleIndex ].iHeight = ( char )( temp->intensity * sin( fSinusSeed + 0.7 * uStimuIndex ) );
		waterplane[ bWhichPlane ][ uParticleIndex ].fSpeed *= 0.4;

		// Next Stimu
		temp = temp->next;
		uStimuIndex++;
	}
}


// *****************************************************************************
void Pool::StepPlane()
{
	Particle * source = waterplane[  bWhichPlane ];
	Particle * dest   = waterplane[ !bWhichPlane ];
		 
	int x;
	int y;
	int i = 0;

	for( y = 0; y < uHeight; y++ )
	{
		for( x = 0; x < uWidth; x++, i++ )
		{
			static int selfIndex;
			selfIndex = i; // y * uWidth + x;

			static signed char selfHeight;
			selfHeight = source[ selfIndex ].iHeight;

			static int diff;
			diff = 0;

			if( x + 1 < uWidth )
			{
				diff += source[ selfIndex + 1 ].iHeight - selfHeight;	// right

				if( y + 1 < uHeight ) 
					diff += source[ selfIndex + uWidth + 1 ].iHeight - selfHeight;	// upper right

				if( y - 1 >= 0 )
					diff += source[ selfIndex - uWidth + 1 ].iHeight - selfHeight;	// lower right
			}

			if( x - 1 >= 0 )
			{
				diff += source [selfIndex - 1].iHeight - selfHeight;	// left

				if( y + 1 < uHeight )
					diff += source[ selfIndex + uWidth - 1 ].iHeight - selfHeight;	// upper left

				if( y - 1 >= 0 )
					diff += source[ selfIndex - uWidth - 1 ].iHeight - selfHeight;	// lower left
			}

			if( y + 1 < uHeight )
				diff += source[ selfIndex + uWidth ].iHeight - selfHeight;	// above

			if( y - 1 >= 0 )
				diff += source[ selfIndex - uWidth ].iHeight - selfHeight;	// below


			float speed = source[ selfIndex ].fSpeed + diff;

			int height = selfHeight + ( int )( speed / 10.0f );
			if( height > 126 )
				height = 126;
			else if( height < -126 )
				height = -126;

			dest[ selfIndex ].fSpeed = speed * 0.972f;
			dest[ selfIndex ].iHeight = ( char )height;
		}
	}
}


// *****************************************************************************
void Pool::UpdateImage( bool bRefreshAll )
{
	const int ZERO_LEVEL = 127;
	const int MAX_HEIGHT = 127;

	UINT screenX;
	UINT screenY;
	int heightFactor;

	UINT uIndex = POOL_WIDTH * POOL_HEIGHT;
	while( uIndex-- )
	{
		if( bRefreshAll || ( waterplane[ 0 ][ uIndex ].iHeight != waterplane[ 1 ][ uIndex ].iHeight ) )
		{
			int x = uIndex % uWidth;
			int y = uIndex / uWidth;
			int height = waterplane[ bWhichPlane ][ uIndex ].iHeight;

			if( height > 40 )
				height = 40;
			else if( height < -40 )
				height = -40;

			heightFactor = 235 + ( ( height - MAX_HEIGHT ) / 2 );

#ifdef USE_VIDEO_FOR_WINDOWS
			screenX = uWidth - x;
			screenY = y;
			image[ uIndex ].rgbtBlue = heightFactor * (screenX + screenY) / (uHeight + uWidth);
			image[ uIndex ].rgbtGreen = heightFactor * screenY / uHeight;
			image[ uIndex ].rgbtRed = height + ZERO_LEVEL;
#else
			screenX = uWidth - x;
			screenY = uHeight - y;
			image[ uIndex ].rgbtRed = heightFactor * (screenX + screenY) / (uHeight + uWidth);
			image[ uIndex ].rgbtGreen = heightFactor * screenY / uHeight;
			image[ uIndex ].rgbtBlue = height + ZERO_LEVEL;
#endif
		}
	}
}



// *****************************************************************************
void Pool::NextFrame()
{
	// Switch plane
	bWhichPlane = !bWhichPlane;

	// Make enough stimus
	if( uStimus < MAX_STIMUS ) AddStimu();

	// Oscillate stimus
	StepStimus();
	
	// Update height field
	StepPlane();
	
	// Update image where modified
	UpdateImage( false );
}
